#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filenames: select.c, select-common.c
# Original copyright and license from select.c:
#/*
#    A system to permit user selection of a block and rotation axis
#    of a magic cube.
#    Copyright (C) 1998,  2003  John Darrington
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/
# Original copyright and license from select-common.c:
#/*
#  A system to permit user selection of a block and rotation axis
#  of a magic cube.
#  Copyright (C) 2003  John Darrington
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


# This library provides  a means of picking a block using the mouse cursor.
# 
# Two mutually co-operative mechanisms are used in this library.  There is a
# timer callback,  which occurs at regular intervals.  There is also the mouse
# motion callback,  which occurs whenever the mouse cursor is moving.  If two
# consecutive timer callbacks occur,  without and intervening mouse motion callback,
# then the cursor is assumed to be stationary.
# 
# If a stationary mouse is detected,  the program goes on to determine which block
# in the cube (if any) the cursor is located upon.


import math

import ui

import gobject
from gtk import gdk

import glarea
import glarea_common
import drwBlock
import cube
import txfm

from OpenGL import GL as gl
from OpenGL import GLU as glu


class FacetSelection:
    def __init__(self, block=0, face=0, quadrant=0):
        self.block = block  # int
        self.face = face    # int
        self.quadrant = quadrant  # int
    def __str__(self):
        return 'FacetSelection(block=%s face=%s quadrant=%s)' % (self.block, self.face, self.quadrant)


idle_threshold = 0
motion = False
Stop_Detected = False


timerActive = False
timer = 0

# is the selection mechanism necessary ? 
needSelection = False

BUFSIZE = 512

mouse_x = -1
mouse_y = -1

granularity = -1
action_select_func = None

noItemSelected = 1

# Structure to hold copy of the last selection taken
current_selection = FacetSelection(-1, -1, -1)


# Initialise the selection mechanism.  Holdoff is the time for which
# the mouse must stay still,  for anything to happen. Precision is the
# minimum distance it must have moved. Do_this is a pointer to a function
# to be called when a new block is selected. 
#@debug
def initSelection (holdoff, precision, do_this):
    global needSelection, idle_threshold, granularity, action_select_func
    needSelection = False
    idle_threshold = holdoff
    granularity = precision
    #print 'holdoff', holdoff, 'precision', precision

    glarea.glwidget.connect ("motion_notify_event", detect_motion, 0)


    action_select_func = do_this

    global timer, timerActive
    timer = gobject.timeout_add (idle_threshold,  UnsetMotion)

    timerActive = True

    # Add a handler to for all those occasions when we don't need the
    # selection mechanism going 

    glarea.glwidget.connect ("enter-notify-event", enableDisableSelection, 0)
    glarea.glwidget.connect ("leave-notify-event", enableDisableSelection, 0)
    glarea.glwidget.connect ("visibility-notify-event", enableDisableSelection, 0)
    glarea.glwidget.connect ("unmap-event", enableDisableSelection, 0)


def disableSelection ():
    global timerActive
    gobject.source_remove (timer)
    timerActive = False


def enableSelection ():
    global timerActive, timer, needSelection
    if not timerActive:
        timer = gobject.timeout_add (idle_threshold,  UnsetMotion)
        timerActive = True

    needSelection = True


# When the window is not mapped,  kill the selection mechanism.  It wastes
# processor time 
enableDisableSelection_entered = 0
def enableDisableSelection (unused_w, event, unused_data):
    # This is a kludge to work around a rather horrible bug;  for some
    # reason,  some  platforms emit a EnterNotify and LeaveNotify (in
    # that order) when animations occur.  This workaround makes sure
    # that the window is not `entered twice' 

    global needSelection, timer, timerActive, enableDisableSelection_entered
    if event.type == gdk.ENTER_NOTIFY:
        enableDisableSelection_entered += 1
        needSelection = False
        if not timerActive:
            timer = gobject.timeout_add (idle_threshold,  UnsetMotion)
            timerActive = True

    elif event.type == gdk.LEAVE_NOTIFY:
        updateSelection ()
        enableDisableSelection_entered -= 1
        if enableDisableSelection_entered <= 0:
            needSelection = True
            gobject.source_remove (timer)
            timerActive = False

    return False



# This callback occurs whenever the mouse is moving 
def detect_motion (unused_w, event, unused_user_data):
    if event.type != gdk.MOTION_NOTIFY:
        return False

    global mouse_x, mouse_y
    mouse_x = event.x
    mouse_y = event.y

    global motion, Stop_Detected
    motion = True
    Stop_Detected = False

    return False



# This callback occurs at regular intervals.   The period is determined by
# idle_threshold.  It checks to see if the mouse has moved,  since the last
# call of this function.
# Post-condition:  motion is False.
def UnsetMotion ():
    global motion, Stop_Detected
    if motion == False: # if not moved since last time 
        if not Stop_Detected:
            # in here,  things happen upon the mouse stopping 
            Stop_Detected = True
            updateSelection()

    motion = False

    return True



def get_widget_height (w):
    return w.allocation.height



# Identify the block at screen co-ordinates x,  y .  This func determines all
# candidate blocks.  That is,  all blocks which orthogonally project to x,  y.  It
# then calls choose_items,  to determine which of them is closest to the screen.
def pickPolygons (x, y):
    #GLint height

    viewport = (0,)*4
    #selectBuf = [0]*BUFSIZE
    #GLint hits

    assert (granularity > 0)
    assert (glarea_common.fovy > 0)
    assert (glarea_common.cp_near > 0)
    assert (glarea_common.cp_far > 0)

    height = get_widget_height(glarea.glwidget)
    #print 'widget_height', height

    gl.glSelectBuffer (BUFSIZE)

    gl.glRenderMode (gl.GL_SELECT)

    gl.glInitNames ()
    gl.glPushName (0xFFFF)

    gl.glMatrixMode (gl.GL_PROJECTION)
    gl.glPushMatrix ()
    gl.glLoadIdentity ()

    viewport = gl.glGetIntegerv (gl.GL_VIEWPORT)

    glu.gluPickMatrix ( x, (height - y), granularity,  granularity,  viewport)

    glu.gluPerspective (glarea_common.fovy,  1,  glarea_common.cp_near,  glarea_common.cp_far)

    glarea_common.modelViewInit ()
    drwBlock.pick_cube()
    gl.glMatrixMode (gl.GL_PROJECTION)
    gl.glPopMatrix ()

    glarea.error_check("")
    #hits = gl.glRenderMode (gl.GL_RENDER)
    selectBuf = gl.glRenderMode (gl.GL_RENDER)
    hits = len(selectBuf)
    #print hits, 'selectBuf', selectBuf and dir(selectBuf[0])
    if hits == 0:
        #print 'hits == 0'
        glarea.error_check("Error Selecting")
        return None

    return choose_items (hits,  selectBuf)



# Find out which of all the objects in buffer is the one with the
# lowest Z value.  ie. closer in the depth buffer
def choose_items (hits, buffer):
    assert hits == len(buffer)
    if not buffer:
        return None
    
    buf = min(buffer, key=lambda x:x.near)
    #buf = min(buffer, key=lambda x:x.far)
    #zvalue = buf.far
    closest = buf.names

    if len(closest) < 3:
        return None
    #SEL_BLOCK = 0
    #SEL_FACE = 1
    #SEL_QUAD = 2
    selection = FacetSelection()
    selection.block = closest[0]
    selection.face = closest[1]
    selection.quadrant = closest[2]
    #print 'selection',selection

    return selection


# an accessor func to get the value of the currently selected items
def selectedItems ():
    if noItemSelected:
        return None

    return current_selection



# This func,  determines which block the mouse is pointing at,  and if it
# has changed,  calls the function ptr "action_select_func"
def updateSelection():
    global noItemSelected, current_selection
    
    selected_polygons = pickPolygons(mouse_x, mouse_y)
    #print noItemSelected, current_selection, selected_polygons
    if selected_polygons == None:
        noItemSelected = 1
    else:
        noItemSelected = 0
        current_selection = selected_polygons

    if action_select_func:
      action_select_func()


def itemIsSelected ():
    return not noItemSelected


# This func is called whenever a new set of polygons have been selected.
# Call redisplay
#@debug
def selection_func():
    #assert False
    if ui.animation_in_progress:
        return

    selection = selectedItems()
    if selection is not None:
        turn_axis = ui.getTurnAxis (selection)
        ui.pending_movement.dir = ui.turnDir(turn_axis)
        ui.pending_movement.axis = ui.vector2axis(turn_axis)
        ui.pending_movement.slice = cube.get_pending_movement_slice(
                                                selection.block,
                                                ui.pending_movement.axis)

        # Here we take the orientation of the selected quadrant and multiply it
        # by the projection matrix.  The result gives us the angle (on the screen)
        # at which the mouse cursor needs to be drawn.
        proj = gl.glGetFloatv (gl.GL_PROJECTION_MATRIX)
        v = cube.get_quadrant_vector(selection.block, selection.face, selection.quadrant)
        v = txfm.transform(proj, v)

        glarea.cursorAngle = math.atan2(v[0], v[1]) * 180.0 / math.pi
        
    else:
        ui.pending_movement.dir = -1
        ui.pending_movement.axis = -1

    #TODO Test this
    #glarea.postRedisplay()
    glarea.postSetCursor()


