#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: ui.c
# Original copyright and license:
#/*
#    User Interface functions for the Cube
#    Copyright (C) 1998,  2003  John Darrington
#                  2004  John Darrington,  Dale Mellor
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License,  or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


import txfm
import cube
import drwBlock
import glarea
import select_
import widget_set

import quarternion
import move_queue_m
import glarea_common
from confstore import confstore

import gtk.keysyms
import gobject


GNUBIK_Left = gtk.keysyms.Left
GNUBIK_Right = gtk.keysyms.Right
GNUBIK_Up  = gtk.keysyms.Up
GNUBIK_Down = gtk.keysyms.Down

abort_requested = 0

# picture rate in ms. 40ms = 25 frames per sec
picture_rate = 40
animation_in_progress = 0
# this is the number of frames drawn when a slice of the cube rotates 90 degrees
frameQty = confstore.frameQty

move_queue = None

# The move taking place NOW
#TODO ?
class _Current_Movement:
    def __init__(self):
        self.move_data = move_queue_m.MoveData()
        self.stop_requested = 0
current_movement = _Current_Movement()


# The move that will take place when the mouse is clicked
pending_movement = move_queue_m.MoveData(-1, -1, -1)


def is_animating():
    return animation_in_progress


# Rotate cube about axis (screen relative)
def rotate_cube(axis, dir_):
    # how many degrees to turn the cube with each hit
    step = 2.0
    rot = quarternion.Quarternion()
    #vector v
    v = txfm.Vector()

    if dir_:
        step = -step

    if axis == 0:
        v[1] = v[2] = 0
        v[0] = 1
    elif axis == 1:
        v[0] = v[2] = 0
        v[1] = 1
    elif axis == 2:
        v[0] = v[1] = 0
        v[2] = 1

    quarternion.quarternion_from_rotation(rot, v, step)
    quarternion.quarternion_pre_mult(glarea_common.cube_view, rot)
    glarea.postRedisplay()


# orientate the whole cube with the arrow keys
def do_keysyms(keysym, shifted):
    if keysym == GNUBIK_Right:
        if shifted:
            dir_ = 0
            axis = 2
        else:
            dir_ = 1
            axis = 1
    elif keysym == GNUBIK_Left:
        if shifted:
            dir_ = 1
            axis = 2
        else:
            dir_ = 0
            axis = 1
    elif keysym == GNUBIK_Up:
        axis = 0
        dir_ = 0
    elif keysym == GNUBIK_Down:
        axis = 0
        dir_ = 1
    else:
        return

    rotate_cube(axis, dir_)
    return True


# return the turn direction,  based upon the turn axis vector
#@debug
def turnDir(vector):
    # This is a horrendous kludge.  It works only because we know that
    # vector is arithemtically very simple

    return vector[0] + vector[1] + vector[2]  >  0




# Convert a vector to a axis number.  Obviously this assumes the vector
#is orthogonal to the frame of reference ( another nasty kludge ).
def vector2axis(vector):
    return (0 if vector[0] != 0 else
            1 if vector[1] != 0 else
            2)




turn_axesX  = [1, 0, 0, 0]
turn_axes_X = [-1, 0, 0, 0]
turn_axesY  = [0, 1, 0, 0]
turn_axes_Y = [0, -1, 0, 0]
turn_axesZ  = [0, 0, 1, 0]
turn_axes_Z = [0, 0, -1, 0]
turn_axes = [
                [turn_axes_Y, turn_axesX,  turn_axesY,  turn_axes_X],
                [turn_axesY,  turn_axesX,  turn_axes_Y, turn_axes_X],
                [turn_axesZ,  turn_axesX,  turn_axes_Z, turn_axes_X],
                [turn_axes_Z, turn_axesX,  turn_axesZ,  turn_axes_X],
                [turn_axes_Y, turn_axes_Z, turn_axesY,  turn_axesZ],
                [turn_axesY,  turn_axes_Z, turn_axes_Y, turn_axesZ],
            ]

# Determine the axis about which to rotate the slice,  from the objects
# selected by the cursor position
#@debug
def getTurnAxis(items):
    # Each edge (quadrant) on a block represents a different axis

    # Fetch the selected block's transformation from its original
    # orientation
    t = cube.get_block_transform(items.block)

    # Select the untransformed vector for the selected edge
    vector = [turn_axes[items.face][items.quadrant][i] for i in xrange(4)]

    # transform it,  so that we go the right way
    vector = txfm.transform(t, vector)
    
    return vector


# handle mouse clicks
def mouse(button, application):
    # Don't let a user make a move,  whilst one is already in progress,
    # otherwise the cube falls to bits.
    if animation_in_progress:
        return

    if button == 1:
        # Make a move
        if select_.itemIsSelected():
            request_rotation(pending_movement, application)
        
        glarea.postRedisplay()


### Funktions to control the cube for use in callbacks and scripts

#@debug
def request_stop():
    '''Stop when the current animation completes'''
    current_movement.stop_requested = 1

def request_abort():
    '''Stop as soon as possible.'''
    global abort_requested
    abort_requested = 1

#@debug
def request_back(application):
    '''One step back in the sequence of moves'''
    if not animation_in_progress and move_queue and move_queue.retard():
        current_movement.stop_requested = 1
        next_move = move_queue.current().data.copy()
        next_move.dir = not next_move.dir
        animate_rotation(next_move, application)

# Internal functionality for the following two functions.
#@debug
def _request_play(one_move, application):
    if not animation_in_progress and move_queue and move_queue.current():
        current_movement.stop_requested = one_move
        next_move = move_queue.current().data.copy()
        move_queue.advance()
        animate_rotation(next_move, application)
        return True
    return False

#@debug
def request_play(application):
    '''Play the current sequence of moves to the end'''
    _request_play(0, application)

#@debug
def request_next(application):
    '''One step forward in the sequence of moves'''
    _request_play(1, application)

#@debug
def request_rotation(data, application):
    '''Make one new move.
    
    The usual rotation request,  called when user uses mouse to rotate the cube by
    hand. The move is put into the `current' place,  and all the moves in the
    move_queue afterwards are zapped.'''
    global move_queue
    if move_queue == None:
        move_queue = move_queue_m.MoveQueue()
    move_queue.push_current(data)
    application.update_statusbar()
    request_play(application)

#@debug
def request_delayed_rotation(data):
    '''Apppend one step to the sequence of moves.
    
    The rotation request called from scripts. In this case the moves are
    appended to the move_queue,  so the script can race ahead and work out the
    moves,  leaving the graphics to catch up later.'''
    global move_queue
    if move_queue == None:
        move_queue = move_queue_m.MoveQueue()
    move_queue.push(data)

#@debug
def request_truncate_move_queue():
    '''Discard moves from current position to the end in the sequence of moves.
    
    Scripts calls this function on startup,  and then the above function as each
    move is computed. This way,  the moves computed are appended to the move_queue
    from the current point forwards (and forward data that was on the queue is
    discarded).'''
    if move_queue:
        move_queue.truncate()

#@debug
def request_mark_move_queue(application=None, mark=True):
    '''Set a marker at the current position in the sequence of moves'''
    if move_queue:
        move_queue.mark_current(mark)
    # if called from scripts update_ui is called later
    if application:
        update_ui(application)

#@debug
def request_queue_rewind(application):
    '''Go to first position in the sequence of moves.
    
    All the moves take place on the cube,  not via the animation
    routines,  so the effect is to unwind the queue instantaneously.'''
    if not move_queue:
        return

    while True:
        mark_result = move_queue.retard()
        if mark_result == 0:
            break
        move_data = move_queue.current().data
        if move_data != None:
            cube.rotate_back(move_data)
        if mark_result != 1:
            break
    glarea.postRedisplay()
    update_ui(application)

#@debug
def request_fast_forward(application=None):
    '''Go to end of the sequence of moves.
    
    All the moves take place on the cube,  not via the animation
    routines,  so the effect is to show the result instantaneously.'''
    if not move_queue:
        return

    move_data = move_queue.current()
    while move_data:
        blocks = cube.identify_blocks_2(move_data.data.slice, move_data.data.axis)
        cube.rotate_slice(3 if move_data.data.dir==0 else 1, blocks)
        move_queue.advance()
        move_data = move_queue.current()
        if move_data is None or move_data.mark:
            break

    glarea.postRedisplay()
    if application:
        update_ui(application)


### Helper functions

@debug
def animate_rotation(data, application):
    assert isinstance(data, move_queue_m.MoveData), str(data)
    assert type(application)==widget_set.MainWindow, application
    
    current_movement.move_data = data.copy()
    drwBlock.start_animate(data)
    global animation_in_progress
    animation_in_progress = 1
    #TODO: need to know mark, but its in move_item not move_data
    application.playbarstate.playing(False)
    #application.playbarstate.playing(move_item.mark)
    gobject.timeout_add(picture_rate, on_animate, current_movement, application)

# a timeout  calls this func,  to animate the cube
#@debug
def on_animate(data, application):
    global abort_requested
    #assert isinstance(data, move_queue_m.MoveData), repr(data)
    md = data
    #assert isinstance(md.move_data, move_queue_m.MoveData), str(md.move_data)

    # how many degrees motion per frame
    increment = 90.0 / (frameQty + 1)

    # decrement the current angle
    unfinished = drwBlock.step_animate(increment)

    # and redraw it
    glarea.postRedisplay ()

    if unfinished and not abort_requested:
        # call this timeout again
        gobject.timeout_add (picture_rate, on_animate, data, application)
    else:
        # we have finished the animation sequence now
        drwBlock.end_animate(md.move_data)

        global animation_in_progress
        animation_in_progress = 0
        
        select_.updateSelection()
        
        if abort_requested:
            abort_requested = 0

        # If there are more animations in the queue,  go again unless a stop has
        # been requested.

        #FIXME ??? md or current_movement ???
        if move_queue.current() and not md.stop_requested:
            application.update_statusbar()
            md.move_data = move_queue.current().data.copy()
            move_queue.advance()
            animate_rotation(md.move_data, application)
        else:
            md.stop_requested = 0
            update_ui(application)
        
    return False

@debug
def update_ui(application):
    assert type(application)==widget_set.MainWindow, application
    
    if move_queue.progress().current:
        if move_queue.current():
            application.playbarstate.mid(move_queue.current().mark)
        else:
            application.playbarstate.last(move_queue.mark_next_push)
    else:
        if move_queue.current():
            application.playbarstate.first()
        else:
            application.playbarstate.empty()
        
    application.update_statusbar()
    
    status = cube.cube_status_check()
    if cube.NOT_SOLVED != status:
        application.declare_win(status)

@debug
def update_ui_first(application):
    assert type(application)==widget_set.MainWindow, application
    
    #TODO move_queue should never be None
    #application.playbarstate.empty()
        
    application.update_statusbar_first()
    
    status = cube.cube_status_check()
    #if cube.NOT_SOLVED != status:
    #    application.declare_win_first(status)

if DEBUG:
    # Manually make a move on the cube
    def on_manual_move(unused_w, data, application):
        # Just loose the first argument and call the next function
        request_rotation(data, application)
        glarea.postRedisplay()

