#!/usr/bin/env python

from distutils.core import setup
from distutils.dist import Distribution
from distutils.cmd import Command
from distutils.command.install_data import install_data
from distutils.command.build import build
from distutils.dep_util import newer
from distutils.log import warn, info, error, fatal

from distutils.extension import Extension
from distutils.command.build_ext import build_ext
#from Pyrex.Distutils.extension import Extension
#from Pyrex.Distutils import build_ext
#from distutils.extension import Extension
#from Cython.Distutils import build_ext

#from DistUtilsExtra.command.build_i18n import build_i18n
#from DistUtilsExtra.command.clean_i18n import clean_i18n

from glob import glob
import os
import sys
import subprocess
import platform

from pybiklib import version

PO_DIR = 'po'
MO_DIR = os.path.join('build', 'mo')


class PybikDist(Distribution):
    global_options = Distribution.global_options + [
        ("without-gettext", None, "Don't build/install gettext .mo files"),
        #("without-icon-cache", None, "Don't attempt to run gtk-update-icon-cache")
        ]

    def __init__ (self, *args):
        self.without_gettext = False
        #self.without_icon_cache = False
        Distribution.__init__(self, *args)


class BuildData(build):
    def run (self):
        build.run (self)

        if self.distribution.without_gettext:
            return

        for po in glob(os.path.join(PO_DIR, '*.po')):
            lang = os.path.basename(po[:-3])
            mo = os.path.join(MO_DIR, lang, 'pybik.mo')

            directory = os.path.dirname(mo)
            if not os.path.exists(directory):
                info('creating %s' % directory)
                os.makedirs(directory)

            if newer(po, mo):
                info('compiling %s -> %s' % (po, mo))
                try:
                    rc = subprocess.call(['msgfmt', '-o', mo, po])
                    if rc != 0:
                        raise Warning, "msgfmt returned %d" % rc
                except Exception, e:
                    error("Building gettext files failed.  Try setup.py --without-gettext [build|install]")
                    error("Error: %s" % str(e))
                    sys.exit(1)

        TOP_BUILDDIR='.'
        INTLTOOL_MERGE='intltool-merge'
        desktop_in='data/pybik.desktop.in'
        desktop_data='data/pybik.desktop'
        os.system ("C_ALL=C " + INTLTOOL_MERGE + " -d -u -c " + TOP_BUILDDIR +
                   "/po/.intltool-merge-cache " + TOP_BUILDDIR + "/po " +
                   desktop_in + " " + desktop_data)


def py2pyx(infile, outfile):
    info('py2pyx: %s --> %s' % (infile, outfile))
    os.system('python tools/py2pyx.py %s %s' % (infile, outfile))
def pyrex(infile, outfile):
    info('pyrex: %s --> %s' % (infile, outfile))
    os.system("pyrexc %s" % infile)
def cython(infile, outfile):
    info('cython: %s --> %s' % (in_file, out_file))
    os.system("cython {0} 2>{0}.lis; grep '^/' {0}.lis".format(in_file))


class BuildExt (build_ext):
    user_options = build_ext.user_options + [('cython',None,'Use Cython instead of pyrex')]
    def __init__(self, *args):
        self.cython = False
        build_ext.__init__(self, *args)

    def swig_sources(self, sources, extension):
        in_files = sources + extension.depends
        pyx_files = []
        pyx_files_dep = []
        for in_file in in_files:
            out_files = pyx_files if in_file in sources else pyx_files_dep
            if in_file.endswith('.py'):
                out_file = os.path.join(self.build_temp, in_file[:-3]+'_c.pyx')
                self.mkpath(os.path.dirname(out_file))
                self.make_file(in_files, out_file, py2pyx, (in_file, out_file))
            elif in_file.endswith(('.pxi','.pxd')):
                out_file = os.path.join(self.build_temp, in_file)
                self.mkpath(os.path.dirname(out_file))
                self.copy_file(in_file, out_file)
            else:
                out_file = in_file
            out_files.append(out_file)
        out_files = []
        for in_file in pyx_files:
            if in_file.endswith('.pyx'):
                out_file = in_file[:-4]+'.c'
                self.make_file(pyx_files+pyx_files_dep, out_file,
                                cython if self.cython else pyrex, (in_file, out_file))
                out_files.append(out_file)
            else:
                out_files.append(in_file)
        return build_ext.swig_sources(self, out_files, extension)


class Uninstall(Command):
    description = "Attempt an uninstall from an install --record file"

    user_options = [('manifest=', None, 'Installation record filename')]

    def initialize_options(self):
        self.manifest = None

    def finalize_options(self):
        pass

    def get_command_name(self):
        return 'uninstall'

    def run(self):
        self.ensure_filename('manifest')
        try:
            try:
                f = open(self.manifest)
                files = [file.strip() for file in f]
            except IOError, e:
                raise DistutilsFileError("unable to open install manifest: %s", str(e))
        finally:
            f.close()

        for file in files:
            if os.path.isfile(file) or os.path.islink(file):
                info("removing %s" % repr(file))
                if not self.dry_run:
                    try:
                        os.unlink(file)
                    except OSError, e:
                        warn("could not delete: %s" % repr(file))
            elif not os.path.isdir(file):
                info("skipping %s" % repr(file))

        dirs = set()
        for file in reversed(sorted(files)):
            dir = os.path.dirname(file)
            if dir not in dirs and os.path.isdir(dir) and len(os.listdir(dir)) == 0:
                dirs.add(dir)
                # Only nuke empty Python library directories, else we could destroy
                # e.g. locale directories we're the only app with a .mo installed for.
                if dir.find("site-packages/") > 0:
                    info("removing %s" % repr(dir))
                    if not self.dry_run:
                        try:
                            os.rmdir(dir)
                        except OSError, e:
                            warn("could not remove directory: %s" % str(e))
                else:
                    info("skipping empty directory %s" % repr(dir))


class InstallData(install_data):
    def run (self):
        self.data_files.extend (self._find_mo_files ())
        install_data.run (self)
        #if not self.distribution.without_icon_cache:
        #    self._update_icon_cache ()

    # We should do this on uninstall too
    def _update_icon_cache(self):
        info("running gtk-update-icon-cache")
        try:
            subprocess.call(["gtk-update-icon-cache", "-q", "-f", "-t", 
                                os.path.join(self.install_dir, "share/icons/hicolor")])
        except Exception, e:
            warn("updating the GTK icon cache failed: %s" % str(e))

    def _find_mo_files (self):
        data_files = []

        if not self.distribution.without_gettext:
            for mo in glob(os.path.join(MO_DIR, '*', 'pybik.mo')):
                lang = os.path.basename(os.path.dirname(mo))
                dest = os.path.join('share', 'locale', lang, 'LC_MESSAGES')
                data_files.append((dest, [mo]))

        return data_files


if platform.system() == 'FreeBSD':
    man_dir = 'man'
else:
    man_dir = 'share/man'


data_files=[
        ('share/applications', ['data/pybik.desktop']),
        #(os.path.join(man_dir, 'man1'), ['doc/pybik.1']),
        ('share/pixmaps', ['data/pybik.png']),
    ]
package_data_files=[
        '*.py',
        'pybik.ui',
        'pybik.png',
    ]

setup(  name=version.PACKAGE,
        version=version.VERSION,
        description='Pybik - the magic cube',
        author=version.AUTHORS[0],
        author_email=version.CONTACT,
        url=version.WEBSITE,
        license=version.LICENSE_NAME,
        scripts=['pybiklib/pybik'],
        data_files=data_files,
        ext_modules=[ 
                Extension("pybiklib/drwBlock_c",
                            ["pybiklib/drwBlock.py"],
                            depends=["pybiklib/cube.py",
                                    #"pybiklib/cube_c_.pxd",
                                    'pybiklib/opengl_gl.pxi',
                                    ],
                            libraries=['GL']),
            ],
        packages=['pybiklib','pybiklib/data'],
        package_dir={'pybiklib': 'pybiklib', 'pybiklib/data': 'data', },
        package_data={'pybiklib/data': package_data_files},
        cmdclass={'build': BuildData, 'build_ext': BuildExt,
                'install_data': InstallData, 'uninstall': Uninstall,
#                'build_i18n': build_i18n
                },
        distclass=PybikDist
     )

